//================================================================================================
// $Id: codegen_seq.cpp 6514 2023-04-03 18:02:07Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "codegen.h"
#include "tinyframe.h"
#include "welltype/wtruntime.h"
#include "wtruntime/types.h"
#include "welltype/stdlib.h"
#include "wtparser/astvisitor.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_typeinfo.h"
#include "wtparser/ast/ast_simultaneoustype.h"
#include "wtparser/ast/ast_basictype.h"
#include "wtparser/ast/ast_seqtype.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"

namespace wtc
{
namespace codegen
{

void CodeGen::GenSeq__copy_loop(wtast::Type* seqtype, wtast::Type* elemtype, wt_u32 offs_src, wt_u32 offs_dst, wt_u32 offs_i_from, wt_u32 offs_i_to, wt_u32 offs_n, wt_u32 add_i_from, wt_u32 add_i_to)
{
	std::string lab_loop = mEngine.GenLabel();
	std::string lab_end  = mEngine.GenLabel();

	mEngine.AddLabel(lab_loop);
	mEngine.MakePVT(offs_i_to);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddJFL(lab_end); // CMP returns 0 if equal

	mEngine.MakePVT(offs_src);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(offs_i_from);
	mEngine.MakeLD(sizeof(wt_u32));
	if(add_i_from)
	{
		if(AddAssert(mProblemHandler, 0==(add_i_from & (~0xFFu))))
		{
			return;
		}

		mEngine.AddSingle(WT_TOMB_OP_ADD_I, add_i_from);
	}
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeLD(elemtype->GetLoadSize());

	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
		mEngine.AddSingle(WT_TOMB_OP_TDR);
	}

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(offs_i_to);
	mEngine.MakeLD(sizeof(wt_u32));
	if(add_i_to)
	{
		if(AddAssert(mProblemHandler, 0==(add_i_to & (~0xFFu))))
		{
			return;
		}

		mEngine.AddSingle(WT_TOMB_OP_ADD_I, add_i_to);
	}
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeST(elemtype->GetLoadSize());

	if(offs_i_to!=offs_i_from)
	{
		GenHelp__incr(offs_i_to);
	}

	GenHelp__incr(offs_i_from);
	mEngine.AddJmpL(lab_loop);
	mEngine.AddLabel(lab_end);
}

void CodeGen::GenHelp__incr(wt_u32 offs)
{
	mEngine.MakePVT(offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
	mEngine.MakePVT(offs);
	mEngine.MakeST(sizeof(wt_u32));
}

void CodeGen::GenHelp__decr(wt_u32 offs)
{
	mEngine.MakePVT(offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
	mEngine.MakePVT(offs);
	mEngine.MakeST(sizeof(wt_u32));
}

void CodeGen::GenHelp__decr_if_nonzero(wt_u32 offs)
{
	mEngine.MakePVT(offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_JF_I, 6);  // JF
	mEngine.MakePVT(offs);          // 1
	mEngine.MakeLD(sizeof(wt_u32)); // 2
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1); // 3
	mEngine.MakePVT(offs);          // 4
	mEngine.MakeST(sizeof(wt_u32)); // 5
}

void CodeGen::GenSeqCtor(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype)
{
	if(!global_label.empty())
	{
		mEngine.AddGlobal(global_label);
		mGeneratedFuncImplNames.insert(global_label);
	}

	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
	mEngine.MakeSEQN(elemtype->GetLoadSize());
}

void CodeGen::GenSeqDtor(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype)
{
	if(!global_label.empty())
	{
		mEngine.AddGlobal(global_label);
		mGeneratedFuncImplNames.insert(global_label);
	}

	if(elemtype->IsObject())
	{
		TinyFrame   tf;
		std::string lab_loop = mEngine.GenLabel();
		std::string lab_end  = mEngine.GenLabel();
		wt_u32      offs_i   = tf.Add(sizeof(wt_u32));
		wt_u32      offs_n   = tf.Add(sizeof(wt_u32));

		mEngine.MakeDVTA(tf.Size());
		mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
		mEngine.AddSingle(WT_TOMB_OP_SEQL);
		mEngine.MakePVT(offs_n);
		mEngine.MakeST(sizeof(wt_u32));

		mEngine.AddLabel(lab_loop);
		mEngine.MakePVT(offs_i);
		mEngine.MakeLD(sizeof(wt_u32));
		mEngine.MakePVT(offs_n);
		mEngine.MakeLD(sizeof(wt_u32));
		mEngine.AddSingle(WT_TOMB_OP_CMP);
		mEngine.AddJFL(lab_end); // CMP returns 0 if equal

		mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
		mEngine.MakePVT(offs_i);
		mEngine.MakeLD(sizeof(wt_u32));
		mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
		mEngine.MakeLD(elemtype->GetLoadSize());

		mEngine.MakeDUP(sizeof(wt_vmptr));
		if(sizeof(wt_vmptr)!=sizeof(wt_u32)) { mEngine.AddSingle(WT_TOMB_OP_BOR); }
		std::string lab_skip = mEngine.GenLabel();
		mEngine.AddJFI(lab_skip);
		GenDestructorCall(elemtype);
		mEngine.AddLabel(lab_skip);

		GenHelp__incr(offs_i);
		mEngine.AddJmpL(lab_loop);
		mEngine.AddLabel(lab_end);
		mEngine.AddSingle(WT_TOMB_OP_DVTF);
	}

	mEngine.AddSingle(WT_TOMB_OP_SEQD);
}

void CodeGen::GenSeqCopy(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype)
{
	if(!global_label.empty())
	{
		mEngine.AddGlobal(global_label);
		mGeneratedFuncImplNames.insert(global_label);
	}

	TinyFrame tf;

	wt_u32 offs_src = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i   = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n   = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 0, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
}

void CodeGen::GenSeq_empty(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		WTT_BOOL==sig->GetRet(0)->GetTypeId()))
	{
		return;
	}

//	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.MakeEXEC(WTMP_SEQEMPTY);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_seqlen(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		WTT_U32==sig->GetRet(0)->GetTypeId()))
	{
		return;
	}

//	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_seqfill(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 2==sig->GetArgc() && 1==sig->GetRetc() &&
		WTT_U32==sig->GetArg(1)->GetType()->GetTypeId() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetRet(0);

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("seqfill", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame    tf;
	std::string  lab_loop  = mEngine.GenLabel();
	std::string  lab_end   = mEngine.GenLabel();
	wt_u32       offs_elem = tf.Add(loadsize);
	wt_u32       offs_seq  = tf.Add(sizeof(wt_vmptr));
	wt_u32       offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32       offs_n    = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(offs_elem);
	mEngine.MakeST(loadsize);
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(loadsize);
	mEngine.MakePVT(offs_seq);
	mEngine.MakeST(sizeof(wt_vmptr));

	mEngine.AddLabel(lab_loop);
	mEngine.MakePVT(offs_i);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddJFL(lab_end); // CMP returns 0 if equal

	mEngine.MakePVT(offs_elem);
	mEngine.MakeLD(loadsize);

	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
		mEngine.AddSingle(WT_TOMB_OP_TDR);
	}

	mEngine.MakePVT(offs_seq);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(offs_i);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeST(loadsize);

	GenHelp__incr(offs_i);
	mEngine.AddJmpL(lab_loop);
	mEngine.AddLabel(lab_end);

	mEngine.MakePVT(offs_seq);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_subseq(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 3==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		WTT_U32==sig->GetArg(1)->GetType()->GetTypeId() &&
		WTT_U32==sig->GetArg(2)->GetType()->GetTypeId() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("subseq", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame    tf;
	std::string  lab_start = mEngine.GenLabel();
	std::string  lab_chk2  = mEngine.GenLabel();
	wt_u32       offs_src  = tf.Add(sizeof(wt_vmptr));
	wt_u32       offs_dst  = tf.Add(sizeof(wt_vmptr));
	wt_u32       offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32       offs_from = tf.Add(sizeof(wt_u32));
	wt_u32       offs_cnt  = tf.Add(sizeof(wt_u32));
	wt_u32       offs_n    = tf.Add(sizeof(wt_u32)); // seqlen(src)

	mEngine.MakeDVTA(tf.Size());
	mEngine.MakePVT(offs_cnt);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(offs_from);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));

	// from<n
	mEngine.MakePVT(offs_from);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGT);
	mEngine.AddJFL(lab_chk2);
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
	mEngine.MakePVT(offs_cnt);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.AddJmpL(lab_start);

	// from+cnt<n
	mEngine.AddLabel(lab_chk2);
	mEngine.MakePVT(offs_from);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(offs_cnt);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_ADD_S);
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGT);
	mEngine.AddJFL(lab_start);
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(offs_from);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SUB_S);
	mEngine.MakePVT(offs_cnt);
	mEngine.MakeST(sizeof(wt_u32));

	mEngine.AddLabel(lab_start);
	mEngine.MakePVT(offs_cnt);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(loadsize);
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_from, offs_i, offs_cnt, 0, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_lov(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("lov", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	mEngine.AddSingle(WT_TOMB_OP_SEQIR_I, 0);
	mEngine.MakeLD(loadsize);

	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
	}

	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_hiv(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("hiv", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1); // don't care about overflow -> IndexOutOfRangeException
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeLD(loadsize);

	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
	}

	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_hirem(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("hirem", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_src = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i   = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n   = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	GenHelp__decr_if_nonzero(offs_n);
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 0, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_lorem(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("lorem", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_src = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i   = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n   = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	GenHelp__decr_if_nonzero(offs_n);
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 1, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_loext(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 2==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("loext", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_elem = tf.Add(loadsize);
	wt_u32 offs_src  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n    = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.MakePVT(offs_elem);
	mEngine.MakeST(loadsize);
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1); // new size is n+1
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	mEngine.MakePVT(offs_elem);
	mEngine.MakeLD(loadsize);
	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
		mEngine.AddSingle(WT_TOMB_OP_TDR);
	}
	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_I, 0);
	mEngine.MakeST(loadsize);

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 0, 1);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_hiext(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 2==sig->GetArgc() && 1==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("hiext", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_elem = tf.Add(loadsize);
	wt_u32 offs_src  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n    = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.MakePVT(offs_elem);
	mEngine.MakeST(loadsize);
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1); // new size is n+1
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	mEngine.MakePVT(offs_elem);
	mEngine.MakeLD(loadsize);
	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
		mEngine.AddSingle(WT_TOMB_OP_TDR);
	}
	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeST(loadsize);

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 0, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_lopop(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 2==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("lopop", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_elem = tf.Add(loadsize);
	wt_u32 offs_src  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n    = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1); // new size is n-1
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));

	mEngine.MakePVT(offs_src);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_I, 0); // raise IndexOutOfRangeException if empty
	mEngine.MakeLD(loadsize);
	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
	}
	mEngine.MakePVT(offs_elem);
	mEngine.MakeST(loadsize);

	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 1, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.MakePVT(offs_elem);
	mEngine.MakeLD(loadsize);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenSeq_hipop(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler, 1==sig->GetArgc() && 2==sig->GetRetc() &&
		sig->GetArg(0)->GetType()->IsSequence() &&
		sig->GetRet(0)->IsSequence()))
	{
		return;
	}

	wtast::Type* seqtype = sig->GetArg(0)->GetType();

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* elemtype  = static_cast<wtast::SeqType*>(seqtype)->GetElementType();
	wt_u32       loadsize  = elemtype->GetLoadSize();

	if(!elemtype->IsObject())
	{
		std::string impl_name = CreateImplName("hipop", loadsize);

		mEngine.AddJmpL(impl_name);

		if(mGeneratedFuncImplNames.end()!=mGeneratedFuncImplNames.find(impl_name))
		{
			return;
		}

		mEngine.AddGlobal(impl_name);
	}

	TinyFrame tf;

	wt_u32 offs_elem = tf.Add(loadsize);
	wt_u32 offs_src  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_dst  = tf.Add(sizeof(wt_vmptr));
	wt_u32 offs_i    = tf.Add(sizeof(wt_u32));
	wt_u32 offs_n    = tf.Add(sizeof(wt_u32));

	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.MakePVT(offs_src);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1); // new size is n-1
	mEngine.MakePVT(offs_n);
	mEngine.MakeST(sizeof(wt_u32));

	mEngine.MakePVT(offs_src);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S); // raise IndexOutOfRangeException if empty
	mEngine.MakeLD(loadsize);
	if(elemtype->IsObject())
	{
		GenCopyCall(elemtype);
	}
	mEngine.MakePVT(offs_elem);
	mEngine.MakeST(loadsize);

	mEngine.MakePVT(offs_n);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakeSEQN(elemtype->GetLoadSize());
	mEngine.MakePVT(offs_dst);
	mEngine.MakeST(sizeof(wt_vmptr));

	GenSeq__copy_loop(seqtype, elemtype, offs_src, offs_dst, offs_i, offs_i, offs_n, 0, 0);

	mEngine.MakePVT(offs_dst);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(seqtype));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.MakePVT(offs_elem);
	mEngine.MakeLD(loadsize);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

} // codegen
} // wtc
